#!/usr/bin/perl
# Copyright 2003 The Board of Trustees of
# The Leland Stanford Junior University. All Rights Reserved.

#Please send comments, questions, and bug reports to Jessica Ebert (jessicas@stanford.edu)


use File::Basename;
use lib "/var/www/cgi-bin/FoldMiner2";
use Scop qw(%num_SSEs %classifications %numdomains $VERSION %foldnames); #has a hash (%classifications) with scop domains as keys and classifications as values
use Expectation qw(%lambda %mu %lambda_local %mu_local); #has hashes %lambda and %mu containing expectation curve parameters for folds
$global=1;

$mv = "/bin/mv";
$mkdir = "/bin/mkdir";
$file = "/usr/bin/file";
$lockprogram = "/usr/local/bin/lock2";
$wc = "/usr/bin/wc";
$gzip = "/bin/gzip";
$gunzip = /bin/gunzip";
############################################################################
# Usage:
#    ./FoldMiner.pl -q <query PDB or SCOP id> -t <full path to target database file> [-a <directory where alignments are stored> -r <full path to search.out file>  -e <expectation value> -x <x value> -exclude <exclude string> -lpg];
# 
#
# Required arguments:
#
# -q   Query structure's PDB or SCOP id: e.g. d1dlwa_ (a SCOP id), 1dlw (a PDB id),
#      or 1dlw-A (pdb id with chain specified)
#      *note:* If the query structure can't be found in local pdb or SCOP 
#              databases specified in site.defs,
#              you must supply the full path to an *uncompressed* pdb file
#
# -t   Full path to the target database file
#      Each line must contain either a full path to a PDB file or the SCOP or PDB 
#      identifier for a target structure. To specify a single chain of a PDB structure, 
#      append a dash and the chain identifier to the PDB identifier (e.g. chain A of 1dlw 
#      becomes 1dlw-A).
#
#
# Optional Arguments:
#
# -r Full path to search.out file (you may rename it),  e.g. "./myresults/run1/search.out"
#   *note:* If not specified, FoldMiner will use the directory supplied with the -a argument,
#           if specified, or the current working directory otherwise 
#
#    
#   *note:* If you would like FoldMiner to run the LOCK 2 alignments, this argument should be the
#           desired location of the search.out file. If you specify a directory that does not exist,
#           it will be created. You may specify a name other than search.out, and FoldMiner will use
#           this name.
#
#   *note:* If you have not already run LOCK 2 alignments, the alignments will be run and placed
#           in the directory (ALIGNMENTDIR) specified in either the configuration file, the directory 
#           specified by the -a command line option, the directory specified by the -r command line
#           option, or the current working directory (options listed in order of precedence).
#
# -a Directory where alignments are/will be stored:
#
#  If you have already run LOCK 2 alignments, give the directory where they are stored.
#  Otherwise, specify the directory where you would like FoldMiner to put the alignments.
#  If excluded, the current working directory will be used
#
# -e Expectation. A value of 10 indicates that, on average, 10 false positives will be reported.
#    If not specified, the default value is 10. 
#
# -x x value: If not specified, the default value is 0.75. The x value
#    determines the extent to which the query secondary structure
#    elements' conservations influence the structural similarity
#    search by giving more weight to highly conserved regions of the
#    query structure. This value must lie between 0 and 1, where a
#    value of 0 turns off the motif detection. A value of 1 (or very
#    close to 1) may lead to undesirable behavior and so is not
#    recommended.
#
# -exclude  Specifies secondary structure elements to be excluded from analysis. Value is a
#           comma separated list of start residues, e.g. 8,25. Optional.
#
# -p Make PDB files for statistically significant alignments
# -l Specifies local (rather than global) alignments
# -g Specifies that alignments should be gzipped, or that they are gzipped 
#    if they've already been run
#####################################################################


&main();



#Reads in target database file
sub readTargetDB {
  my($filename) = @_;
  my($line,@fields);
  my(@targets,$target);
  

  $#targets = -1;
  open(DB, $filename) || die "can't open target database file ($filename): $!\n";
  while($line = <DB>) {
    next if($line =~ /LOCK_LIST/);
    chomp($line);
    if($line =~ m|/(.*)$|) {
      $target = $1;
    }else {
      $target = $line;
    }
    $targets[$#targets+1] = $target;
  }
  return @targets;
}



#Parses the exclude string from configuration file to determine which SSEs, if any, to exclude
sub parseExcludeString {
  my($excludestring) = @_;
  my(@fields,$SSE);
  my(%excluded);


  %excluded = ();
  $exclude_SSEs = 0;
  @fields = split(/\s+/,$excludestring);
  for $SSE (@fields) {
    $excluded{$SSE} = 1;
    $exclude_SSEs++;
  }
  @keys = keys(%excluded);
  return %excluded;
}



#get SCOP classification of the query, if available
sub getSCOPClassification {
  my($query) = @_;
  my($queryclass,$queryfold,$querysuperfam,$queryfam);


  if(!defined($classifications{$query})) {
      if($query =~ /^(.*)(\..*)$/) {
          $query  = $1;
      }
  }

  if(defined($classifications{$query})) {
    ($queryclass,$queryfold,$querysuperfam,$queryfam) = split(/\./,$classifications{$query});
      
    #expectation values are hashed in the form of <class>.<fold>
    $queryfold = $queryclass.".".$queryfold;
    
    # if we do not have a background distribution for this query, 
    # use the overall distribution for all of SCOP
    if(!defined($lambda{$queryfold})) {
      #need to assign background distributions based on class, if possible, 
      # or overall distribution if this is not possible
      if($queryclass eq "a") {
	$queryfold = "a";
      }elsif($queryclass eq "b") {
	$queryfold = "b";
      }elsif($queryclass eq "c") {
	$queryfold = "c";
      }elsif($queryclass eq "d") {
	$queryfold = "d";
      }
      $queryfold = "all"; #always use overall distribution for now
    }
  }else {
    $queryfold = "all";
  }
  return $queryfold;
}

#reads the configuration file, which is named in the parameter passed to this function, 
#and returns parameters for this run
sub readConfigurationFile {
  my($filename) = @_;
  my($line,@fields);
  my($gzipped,$targetdb,$num_false_positives,$xval,$alignment_dir,$excludestring,$make_pdb_files,$query_chain);

  $targetdb = "";
  $alignment_dir="";
  $num_false_positives = -1;
  $xval = -1;
  $query = "";
  $gzipped = 0;
  $excludestring = "";
  $make_pdb_files = 0;
  $query_chain = "";
  open(IN, $filename) || die "can't open configuration file ($filename): $!\n";
  while($line = <IN>) {
    next if($line =~ /^\#/); #ignore comments

    #Check to see if line has a comment, but whole line is not commented out
    if($line =~/(.*[\'\"].*[\'\"]\s*)(\#.*)?/) { 
      $line = $1;
    }elsif($line =~ /^(.+)\#/) {
      $line = $1;
    }


    if($line =~ /^TARGETDB\s*=\s*(.+)\s*$/) {
      $targetdb = $1;
      $targetdb =~ s/[\"\'\s]//g; #remove quotes and whitespace
    }elsif($line =~ /^EXPECTATION\s*=\s*(.+)\s*$/) {
      $num_false_positives = $1;
      $num_false_positives =~ s/[\"\'\s]//g; #remove quotes and whitespace
      if($num_false_positives !~ /^([0-9]+\.)?[0-9]*$/) {
	print STDERR "EXPECTATION parameter in $filename is invalid\n";
	exit(0);
      }
    }elsif($line =~ /^XVAL\s*=\s*(.+)\s*$/) {
      $xval = $1;
      $xval =~ s/[\"\'\s]//g; #remove quotes and whitespace
      if($xval !~ /^([10])?(\.[0-9]*)?$/) {
	print STDERR "XVAL paramter in $filename is invalid\n";exit(0)
      }
      if($xval <0 || $xval > 1) {
	print STDERR "XVAL paramter in $filename is invalid\n";exit(0);
      }
    }elsif($line=~/^EXCLUDE\s*=(.+)\s*$/) {
      $excludestring = $1;
      $excludestring =~ s/[\"\']//g; #remove quotes
      $excludestring =~ s/\s*$//g; #remove trailing whitespace
      $excludestring =~ s/,/ /g; #change commas to spaces
    }elsif($line =~/^MAKE_PDB_FILES/) {
      if($line =~ /1$/) {
	$make_pdb_files = 1;
      }
#      print "make pdb files: $make_pdb_files\n";
    }elsif($line =~ /^GZIPPED\s*=\s*(.+)\s*$/) {
      $gzipped = $1;
    }elsif($line =~ /^ALIGNMENTDIR\s*=\s*(.+)\s*$/) {
      $alignment_dir=$1;
      $alignment_dir =~ s/[\"\'\s]//g; #remove quotes and whitespace
    }elsif($line =~ /^QUERY_CHAIN\s*=\s*(.+)\s*$/) {
        $query_chain=$1;
        $query_chain =~ s/[\"\'\s]//g; #remove quotes and whitespace
   }

    
  }

  #check to make sure all parameters were collected
#  if($targetdb eq "") {
#    print STDERR "TARGETDB not specified in $filename\n";exit(0);
#  }
#  if($num_false_positives == -1) {
#    print STDERR "NUM_FALSE_POSITIVES not specified in $filename\n";exit(0);
#  }
#  if($xval == -1) {
#    print STDERR "XVAL not specified in $filename\n";exit(0);
#  }


  return($targetdb,$num_false_positives,$xval,$excludestring,$make_pdb_files,$gzipped,$alignment_dir,$query_chain);
}


sub readSearchFile {
  my($query,$resultsfile,$resultsdir,$alignment_dir) = @_;
  my($results_dir,$line,$line2,@fields,$i,$start,$start2,$end,$end2,$score,$sstype);
  my($target,$count);
  my($SSEscorefieldstart);
  my($alignment_file);
  my($file);
  my($tclass,$tfold,$tsuperfam,$tfam);

  
  %num_aligned_SSEs = ();
  %finalscores = ();
#  %dp_scores_significant = ();
  %SSE_scores = ();
  %num_aligned_residues = ();
  %RMSD = ();
  %description = (); 


  #test to see if resultsfile has more than one line. If not, and if .out files are present,
  #may need to recreate search.out. If .out files are not present, need to run alignments?

  $SSEscorefieldstart = 6;
  open(IN, $resultsfile) || die "can't open results file $resultsfile: $!\n";
  while($line = <IN>) {
    last if(length($line) < 4); #shouldn't need this check?
    chomp($line);
    @fields = split(/\|/,$line);
    $target = $fields[0];
    if(defined($classifications{$target})) {
	($tclass,$tfold,$tsuperfam,$tfam) = split(/\./,$classifications{$target});
    }else {
	$tclass = "";
    }

    next if($tclass !~ /[abcd]/) && $tclass ne "";
    #need to read in query_target.out to get number of SSEs in target file
#    $alignment_file = $alignment_dir."/".$query."_".$target.".out";
    $alignment_file = $alignment_dir."/query_".$target.".out";

    if($gzipped) {
      $alignment_file .= ".gz";
    }
    if(! -e($alignment_file)) {
      $alignment_file = $alignment_dir."/".$query."_".$target.".out";
      if($gzipped) {
	$alignment_file .= ".gz";
      }
    }

    if($gzipped) {
      open(ALIGNMENT,"zcat -f $alignment_file |") || die "can't open alignment file: $alignment_file: $!\n";
    }else {
      open(ALIGNMENT,$alignment_file) || die "can't open alignment file: $alignment_file: $!\n";
    }
    while($line2 = <ALIGNMENT>) {
      last if($line2 =~ /SSE aa numbers/);
    }
    while($line2 = <ALIGNMENT>) {
      last if($line2 =~/^target/);
    }
    $count = 0;
    while($line2 = <ALIGNMENT>) {
      last if($line2 =~/^CPU/);
      $count++;
    }


    close(ALIGNMENT);
    $num_SSEs{$target} = $count;

    #strip file extension off of otarget
#    if($target =~ /^(.*)(\..*)$/) {
#      $target = $1;
#    }
    my($num_fields) = $#fields-$SSEscorefieldstart+1;
    my($qstart,$qend,@fields2,$tstart,$tend,@ssefields);
    for($i=$SSEscorefieldstart;$i<=$#fields;$i++) {
      ($start,$start2,$end,$end2,$score,$sstype) = split(/,\s*/,$fields[$i]);
      $SSE_scores{$target}{$start} = $score;
    }
    
    $num_aligned_SSEs{$target} = $fields[2];
    $num_aligned_residues{$target} = $fields[3];
    $RMSD{$target} = $fields[4];
    $description{$target} = $fields[5];
    $dp_scores_significant{$target} = $finalscores{$target} = $fields[1];
#    print "processed $target\n";
  }

  close(IN);

  #read query SSE information
  $#query_SSE_starts = -1;
  $#query_SSE_ends = -1;
  $#query_SSE_types = -1;
  my($query_SSEs) = 0;
  my($query_alignment_file) = $alignment_dir."/query_".$query.".out";
  if($gzipped) {
    $query_alignment_file .= ".gz";
  }
  if(! -e($query_alignment_file)) {
    $query_alignment_file = $alignment_dir."/".$query."_".$query.".out";
    if($gzipped) {
      $query_alignment_file .= ".gz";
    }
  }
  if($gzipped) {
    open(ALIGNMENT,"zcat -f $query_alignment_file |") || die "can't open alignment file: $alignment_file: $!\n";
  }else {
    open(ALIGNMENT,$query_alignment_file) || die "can't open query alignment file: $query_alignment_file: $!\n";
  }
  while($line = <ALIGNMENT>) {
    last if($line =~ /^SSE aa numbers/);
  }
  <ALIGNMENT>; #skip line
  while(($line = <ALIGNMENT>) !~ /^target/) {
    ($query_SSE_types[$query_SSEs],$query_SSE_starts[$query_SSEs],$query_SSE_ends[$query_SSEs]) = split(/\s+/,$line);
    $query_SSEs++;
  }
  close(ALIGNMENT);
  return($query_SSEs);
}

sub runAlignments {
  my($query,$targetdb,$targetsref,$alignment_dir,$resultsdir,$resultsfile,$query_chain_opt,$gzipped) = @_;
  my($target,$file);

#  system("$lockprogram -q $query -t $targetdb -o $alignment_dir
#  $query_chain_opt > foldminer.log.tmp");
  system("$lockprogram -q $query -t .targetdb -o $alignment_dir $query_chain_opt > foldminer.log.tmp");
  system("rm -f foldminer.log.tmp");
  if(! -e($resultsdir)) {
    system("$mkdir $resultsdir");
  }
  my($tmpfile) = "$alignment_dir/"."search.out";
  system("$mv $tmpfile $resultsfile");
  if($gzipped) {
    foreach $target (${@targetsref}) {
      $file = $alignment_dir."/$query"."_".$target.".out";
      if(-e $file) {
	system("$gzip -f $file");
      }
    }
  }
}


sub find_conserved_SSEs {
  my($xval,$stringent_pcutoff, $pcutoff,$query_SSEs,$resultsdir)= @_;
  my($i,$j);
  my(@orderbyscore1);
  my($l);
  my($last_index);
  my($type);
  my($line,@fields);
  my($start,$end,$start2,$end2,$temp,$num_results,$num_conserved,$conserved_total);
  my($total_contribution,$target,$file, $scorefield, @SSEs, %scores);
  my($SSE);


  #sort hits by score: problem here?
  @mykeys = keys(%dp_scores_significant);
  @orderbyscore1 = sort(sort_by_score keys(%dp_scores_significant));
  if($#orderbyscore1==-1) { #no statistically significant results
    return(0,@orderbyscore1);
  }

  #find all hits that meet the stringent statistical significance cutoff
  $last_index = $#orderbyscore1;

  for($i=0;$i<=$#orderbyscore1;$i++) {
    if($pvalues{$orderbyscore1[$i]} > $stringent_pcutoff)  {
      $last_index = $i-1;
      last;
    }
  }

  #get lambda such that lowest scoring target contributes 0.01
  #to new maximum score
  if($last_index != $#orderbyscore1) { 
    $l = inverse_exponential_function(($#orderbyscore1-$last_index), 0.01);
  }elsif($#orderbyscore1 != 0) {
    $l = inverse_exponential_function($#orderbyscore1,0.01); #if no hits made stringent cutoff
  }


  %aligned_SSE_counts = ();
  %conserved_query_SSEs = ();
  %total_contributions = ();
  for($i=0;$i<=$#query_SSE_starts;$i++) {
    $aligned_SSE_counts{$query_SSE_starts[$i]} = 0;
    $conserved_query_SSEs{$query_SSE_starts[$i]} = 0; #not conserved by default
    $total_contributions{$query_SSE_starts[$i]} = 0;
  }
  
  
  #find aligned SSEs in the significant results
  $total_contribution = 0;
  for($i=0;$i<=$#orderbyscore1;$i++) {
    $target = $orderbyscore1[$i];
    if($i>$last_index) {
      $temp = exp(-$l*($i-$last_index));
    }else {
      $temp = 1;
    }
    $total_contribution += $temp;

    #need to extract keys from %SSE_scores{$target} 
    %scores = %{$SSE_scores{$target}}; #aligned SSEs for this target are keys, scores are values
    @SSEs = keys(%scores);
    %aligned = ();
    for $SSE (@SSEs) {
      $aligned{$SSE} = 1;
    }

    $found_alignment_start = 0;
    for($j=0;$j<=$#query_SSE_starts;$j++) {
      $SSE = $query_SSE_starts[$j];
      next if(defined($excluded{$SSE}));
      if($j==0) {
	$total_contributions{$SSE} += 20*$temp; #first aligned SSE
	$temp3 = 20*$temp;
	if(defined($aligned{$SSE})) {
	  $found_alignment_start = 1;
	}
      }else {

	if(!$found_alignment_start && defined($aligned{$SSE})) {
	  $temp3 = 20*$temp;
	  $total_contributions{$SSE} += 20*$temp;
	  $found_alignment_start = 1;
	}else {
	  $temp3 = 40*$temp;
	  $total_contributions{$SSE} += 40*$temp;
	}
      }
      $aligned_SSE_counts{$SSE} += $scores{$SSE}*$temp;
    }
    
  }

  #select conserved SSEs
  $num_results = $#orderbyscore1+1;
  $num_conserved = 0;
  $refined_max_score = 0;
  
  $conserved_total = 0;
  for($i=0;$i<$query_SSEs;$i++) {
    if($excluded{$query_SSE_starts[$i]}==1) {
      next;
    }
    if($total_contributions{$query_SSE_starts[$i]} != 0) {
	$temp = $aligned_SSE_counts{$query_SSE_starts[$i]}/$total_contributions{$query_SSE_starts[$i]}*$xval;

	$conserved_query_SSEs{$query_SSE_starts[$i]} = (1-$xval) + $xval*($aligned_SSE_counts{$query_SSE_starts[$i]}/$total_contributions{$query_SSE_starts[$i]});

	$conservation{$query_SSE_starts[$i]} = ($aligned_SSE_counts{$query_SSE_starts[$i]}/$total_contributions{$query_SSE_starts[$i]});
    }else {
      $conserved_query_SSEs{$query_SSE_starts[$i]} = 1-$xval;
      $conservation{$query_SSE_starts[$i]} = 0;
    }    
    if($i==0) {
      $refined_max_score += 20*$conserved_query_SSEs{$query_SSE_starts[$i]};
    }else {
      $refined_max_score += 40*$conserved_query_SSEs{$query_SSE_starts[$i]};
    }
  }
  

  return(($refined_max_score,@orderbyscore1));
}



sub run_query {
  my($query,$query_SSEs,$querymax,$globalquerymax,$use_conserved, $expectation_curve,$pcutoff) = @_;
  my($line,$file,@fields);
  my($analysis_file);
  my($query_target_size_ratio);
  my($targetmax,$target,@targets,$score,%scores,@SSEs, $SSE,$pval);
  my($newquerymax,$newtargetmax);


#  %dp_scores_significant = {};
#  %dp_scores_insignificant = {};
#  %pvalues = {};

  return if($querymax==0);
  return if($query_SSEs < 1);
  @targets = keys(%SSE_scores); #aligned targets

  foreach $target (@targets) {
      if(defined($classifications{$target})) {
	  next if($classifications{$target} !~ /^[abcd]/); #use only four main classes
      }


    %scores = %{$SSE_scores{$target}}; #aligned SSEs for this target are keys, scores are values
    @SSEs = keys(%scores);
    @SSEs = sort {$a <=> $b} @SSEs;
    #find the score for this target
    $score = 0;
    if(!defined($maxscores{$target})) {
      $maxscores{$target} = 40*$num_SSEs{$target}-20;
    }
    if(!$use_conserved) {
      $targetmax = $maxscores{$target};
      $newquerymax = $globalquerymax;
      for($i=0;$i<=$#SSEs;$i++) {
        $SSE = $SSEs[$i];
        next if($excluded{$SSE} == 1);
        $score += $scores{$SSE};
      }
    }else {
      $targetmax = $maxscores{$target};
      $newquerymax = 0;
      $newtargetmax = $targetmax;
      for($i=0;$i<=$#SSEs;$i++) {
        $SSE = $SSEs[$i];
        next if($excluded{$SSE}==1);


        #each SSE score is weighted by its conservation, which is given by $conserved_query_SSEs{$SSE}
        $score += $scores{$SSE}*$conserved_query_SSEs{$SSE};
        if($i==0) {
          $newtargetmax -= (20-$conserved_query_SSEs{$SSE}*20);
        }else {
          $newtargetmax -= (40-$conserved_query_SSEs{$SSE}*40);
        }
      }
    }

    next if($targetmax ==0 && $querymax == 0);
    #    next if(scalar(@SSEs) < 3); #ignore target if too few SSEs are aligned

    # Normalization is based on number of SSEs in the larger of the query and target
    # structures in order to favor alignments that are global with respect to both
    # the query and target. If the query is larger, the score is normalized to the current
    # query vs. query alignment score. If the target is larger, we multiply the original
    # target vs. target alignment score by the ratio of the current maximum query vs. query
    # alignment score to the original query vs. query alignment score. 

    if($num_SSEs{$target} > $query_SSEs) {
	if($global) {
	    $score = $score/($maxscores{$target}*$querymax/$globalquerymax);
	}else {
	    $score = $score/$querymax;
	}
    }else {
	if($global) {
	    $score = $score/$querymax;
	}else {
	    if($newtargetmax != 0) {
		$score = $score/$newtargetmax;
	    }else {
		$score = 0;
	    }
	}
    }

    $finalscores{$target} = $score;
    if($score > 0) {
      if($global) {
	$pval = &expectation_function($lambda{$expectation_curve},$mu{$expectation_curve},$score);
      }else {
	$pval = &expectation_function($lambda_local{$expectation_curve},$mu_local{$expectation_curve},$score);
      }
      $pvalues{$target} = $pval;

      if($pval <= $pcutoff) {
	$dp_scores_significant{$target} = $score;

      }else {
	$dp_scores_insignificant{$target} = $score;
      }
    }else {
      $pvalues{$target} = 1;
    }
  }

}

sub runMotifDiscovery {
  my($query,$query_SSEs,$xval,$expectation_curve,$pcutoff,$stringent_pcutoff,$resultsdir) = @_;
  my($converged) = 0;
  my($globalquerymax);
  my(@hits,$old_score,$old_score_two_rounds,$refined_max_score);

  $globalquerymax = 40*($query_SSEs-$exclude_SSEs)-20; #maximum alignment score
  %dp_scores_significant = ();
  %dp_scores_insignificant = ();
  %pvalues = ();
  run_query($query,$query_SSEs,$globalquerymax,$globalquerymax,0,$expectation_curve,$pcutoff);

  @hits = ();
 
  $old_score = $old_score_two_rounds = $globalquerymax; #initialization
  while(!$converged) {
    ($refined_max_score,@hits) = find_conserved_SSEs($xval,$stringent_pcutoff,$pcutoff,$query_SSEs,$resultsdir);

    @newhits = sort(@hits);

    #check to see if there are no hits
    if($#hits == -1) {
      $converged=0;
      for($i=0;$i<=$#query_SSE_starts;$i++) {
	$conservation{$query_SSE_starts[$i]} = 1;
      }
      last;
    }

    #check for convergence
    if($refined_max_score < $old_score) {
      if($refined_max_score + 2 >= $old_score) {
	$converged = 1;
	last;
      }
    }elsif ($refined_max_score - 2 <= $old_score){
      $converged = 1;
      last;
    }

    #check to see if we're alternating between two maximum scores (happened to me once)
    if($old_score_two_rounds == $refined_max_score) {
      $converged = 1;
      last;
    }
    $old_score_two_rounds = $old_score;
    $old_score = $refined_max_score;

    $old_score = $refined_max_score;
    &run_query($query,$query_SSEs,$refined_max_score,$globalquerymax,1, $expectation_curve,$pcutoff);
  }

  return(@hits);
}
  
sub parseCommandLineOptions {
    my($argsref) = @_;
    my($query,$alignment_dir,$resultsfile,$configuration_file); #options
    my($targetdb,$expectation,$gzipped,$xval,$exclude,$make_pdb_files);
    my($flag,$numargs);
    my($i,$j);
    my(@options);

    $xval = 0.75;
    $expectation = 10;
    $gzipped = 0;
    $make_pdb_files = 0;
    $exclude=$xval=$expectation=$query=$alignment_dir=$resultsfile=$configuration_file="";
    $numargs = scalar(@$argsref);
    for($i=0;$i<$numargs;$i++) {
      if($$argsref[$i] eq "-t") {
        if($i<$numargs-1) {
	  $targetdb = $$argsref[$i+1];
        }else {
	  printUsage();
	  exit(-1);
        }
      }elsif($$argsref[$i] eq "-exclude") {
	if($i < $numargs-1) {
	  $exclude = $$argsref[$i+1]; 
	}
	$exclude =~ s/,/ /g; #change commas to spaces
	
      }elsif($$argsref[$i] eq "-x" || $$argsref[$i] eq "-xval") {
	if($i< $numargs-1) {
	  $xval = $$argsref[$i+1];
	}else {
	  printUsage();
	  exit(-1);
	}
	if($xval !~ /^([10])?(\.[0-9]*)?$/) {
	  print STDERR "XVAL paramter in $filename is invalid\n";exit(0)
	}
	if($xval <0 || $xval > 1) {
	  print STDERR "XVAL paramter in $filename is invalid\n";exit(0);
	}
      }elsif($$argsref[$i] eq "-q") {
	if($i< $numargs-1) {
	  $query = $$argsref[$i+1];
	}else {
	  printUsage();
	  exit(-1);
	}
      }elsif($$argsref[$i] eq "-r" || $$argsref[$i] eq "-resultsdir" || $$argsref[$i] eq "-resultsdirectory" || $$argsref[$i] eq "-resultsfile" || $$argsref[$i] eq "-results") {
	if($i< $numargs-1) {
	  $resultsfile = $$argsref[$i+1];
	}else {
	  printUsage();
	  exit(-1);
	}
      }elsif($$argsref[$i] eq "-a" || $$argsref[$i] eq "-alignments" || $$argsref[$i] eq "-alignment" || $$argsref[$i] eq "-alignmentdir" || $$argsref[$i] eq "-alignmentsdir" || $$argsref[$i] eq "-alignmentdirecotry" || $$argsref[$i] eq "-alignmentsdirectory") {
	if($i < $numargs-1) {
	  $alignment_dir = $$argsref[$i+1];
	}else {
	  printUsage();
	  exit(-1);
	}
      }elsif($$argsref[$i] eq "-c" || $$argsref[$i] eq "-config" || $$argsref[$i] eq "-configfile" || $$argsref[$i] eq "-configuration" || $$argsref[$i] eq "-configurationfile") {
	if($i < $numargs-1) {
	  $configuration_file = $$argsref[$i+1];
	}else {
	  printUsage();
	  exit(-1);
	}
      }elsif($$argsref[$i] eq "-e" || $$argsref[$i] eq "-expectation") {
	if($i<$numargs-1) {
	  if($$argsref[$i+1] =~ /^([0-9]+\.)?[0-9]*$/) {
	    $expectation = $$argsref[$i+1];
	  }elsif($$argsref[$i+1] =~ /^([0-9]+)[eE](-?[0-9]+)$/) {
	    $expectation = $1*(10**$2);
	  }else {
	    print STDERR "expectation parameter ($expectation) is invalid\n";
	    exit(-1);
	  }
	}else {
	  printUsage(); exit(-1);
	}
      }elsif($$argsref[$i] eq "-local") {
	$global = 0;
      }elsif($$argsref[$i] eq "-gzipped") {
	$gzipped = 1;
      }elsif($$argsref[$i] eq "-pdb") {
	$make_pdb_files = 1;
      }elsif($$argsref[$i] =~ /^-([lgp]*)$/) {
	@options = split(//,$1);
	for($j=0;$j<=$#options;$j++) {
	  if($options[$j] eq "g") {
	    $gzipped = 1;
	  }elsif($options[$j] eq "l") {
	    $global = 0;
	  }elsif($options[$j] eq "p") {
	    $make_pdb_files = 1;
	  }else {
	    printUsage();
	    exit(-1);
	  }

	}
      }
    }

    if($query eq "") {
      print STDERR "No query structure specified\n";
      printUsage(-1);
      exit(-1);
    }
    if($targetdb eq "") {
      print STDERR "No target database specified\n";
      printUsage(-1);
      exit(-1);
    }

    return($query,$resultsfile,$alignment_dir,$targetdb,$gzipped,$expectation,$xval,$exclude,$make_pdb_files);
    
}


sub main {
  my($query,$resultsfile,$configurationfile); #arguments for script
  my($targetdb,$num_false_positives,$xval,$alignment_dir,$excludestring,$make_pdb_files,$query_chain); #parameters in config file
  my($alignment_dir_commandline,$gzipped_commandline,$num_false_positives_commandline,$xval_commandline);
  my(@targets,$num_alignments,$num_query_SSEs); 
  my($expectation_curve); #denotes the background distribution used to evaluate statistical significance of alignments
  my(%exclude); #first residue numbers of SSEs to be excluded
  my($pcutoff,$stringent_pcutoff,$resultsdir);
  my($config_alignment_dir);  
  my($query_chain_opt);
  my($target_chain_opt);
  my($query_out_name);
  my($targetdb_commandline);


  $xval=$num_false_positives=$alignment_dir=$gzipped=$num_false_positives=$targetdb = "";
  #check usage
  ($query,$resultsfile,$alignment_dir,$targetdb,$gzipped,$num_false_positives,$xval,$excludestring,$make_pdb_files) = parseCommandLineOptions(\@ARGV);
  print "excludestring: $excludestring\n";
  if($query eq "") {
      print "Expected query\n";
      printUsage();
      exit(-1);
  }

  if($resultsfile eq "") {
      if($alignment_dir ne "") {
        if($alignment_dir =~ /^.*\/$/) {
          chop($alignment_dir);
        }
        $resultsfile = $alignment_dir."/search.out";
       }else {
	 $resultsfile = "./";
       }
  }
  if($alignment_dir eq "") {
     if($resultsfile ne "") {
       ($shortresultsfile,$resultsdir,$suffix) = fileparse($resultsfile);
	$alignment_dir = $resultsdir; 	
     }else {
	 $alignment_dir = "./";
     }
  }

  #read target database
  @targets = readTargetDB($targetdb);
  open(OUT, ">.targetdb");
  print OUT "LOCK_LIST\n";
  foreach $target (@targets) {
      print OUT "$target\n";
  }
  close(OUT);
  #get arguments
  if($query =~ /^\/?([^\/]*\/)*([^\/]+)$/) {
     $shortquery = $2;
   }else {
      $shortquery = $query;
    }
  if($shortquery =~ /^(.*)-(.)$/) {
#    $shortquery = $1;
    $query_chain = $2;
  }else {
    $query_chain = "";
  }
#  print "shortquery: $shortquery\n";

  if($query ne $shortquery) {
      $shortquery = "query";
  }

#  $query_chain = $config_alignment_dir = "";
#  if($configurationfile ne "") {
#    ($targetdb,$num_false_positives,$xval,$excludestring,$make_pdb_files,$gzipped,$config_alignment_dir,$query_chain) = readConfigurationFile($configurationfile);
#  }
#  if($targetdb_commandline ne "") {
#    $targetdb = $targetdb_commandline;
#  } 
#  if($alignment_dir_commandline ne "") {
#    $alignment_dir = $alignment_dir_commandline;
#  }
#  if($num_false_positives_commandline ne "") {
#    $num_false_positives = $num_false_positives_commandline;
#  }
#  if($gzipped_commandline ne "") {
#    $gzipped = $gzipped_commandline;
#  }
#  if($xval_commandline ne "") {
#    $xval = $xval_commandline;
#  }elsif($xval == "") {
#    $xval = 0.75;
#  }
  if($query_chain ne "") {
	$query_chain_opt = "-qc $query_chain";
        $target_chain_opt = "-tc $target_chain";
    }else {
        $query_chain_opt = "";
        $target_chain_opt = "";
    $query_out_name = $query;
   } 
  if($alignment_dir ne "") {
    if($alignment_dir =~ /^.*\/$/) {
      chop($alignment_dir);
    }
    $resultsfile = $alignment_dir."/search.out";
  }


  #check to see if alignments need to be run
  #resultsdir is resultsfile with everything after last "/" stripped off

  ($shortresultsfile,$resultsdir,$suffix) = fileparse($resultsfile);

  my($result,$grepcommand);
  if(-e $resultsfile) {
    $result = `$wc -l $resultsfile`;
    @fields = split(/\s+/,$result);
    if($fields[1] == 1) { #if only one line is present, this is likely not the correct search.out file
      $grepcommand = "ls $alignment_dir | grep -e'_*\.out'";
      $result = `$grepcommand`;
      @files = split("\n",$result);
      $found_alignment_file = 0;
      for $file (@files) {
        if($file =~ /query_.*\.out/ && $file !~ /query_$shortquery\.out/ && $file !~ /$shortquery_$shortquery\.out/ && $file !~ /$shortresultsfile/) {
          $found_alignment_file = 1;
          break;
        }elsif($file =~ /$shortquery_.*\.out/ && $file !~ /query_$shortquery\.out/ && $file !~ /$shortquery_$shortquery\.out/ && $file !~ /$shortresultsfile/) {
          $found_alignment_file = 1;
          break;
        }elsif($file =~ /$query_out_name.*\.out/ && $file !~ /query_$query_out_name\.out/ && $file !~ /$query_out_name_$query_out_name\.out/ && $file !~ /$shortresultsfile/) {
          $found_alignment_file = 1;
          break;
	}
      }
      if(!$found_alignment_file) {
	if(! -e($alignment_dir)) {
	  system("$mkdir $alignment_dir");
	}
	#	print "didn't find alignment files: running alignments\n";
	runAlignments($query,$targetdb,\@targets,$alignment_dir,$resultsdir,$resultsfile,$query_chain_opt,$gzipped);
	
      }
      #      print "found alignment files\n";
    }
  }else {
#    print "didn't find $resultsfile: assuming alignments need to be run\n";
    if(! -e($alignment_dir)) {
      print STDERR "alignment dir ($alignment_dir) doesn't exist: creating\n";
      system("$mkdir $alignment_dir");
    }
    runAlignments($query,$targetdb,\@targets,$alignment_dir,$resultsdir,$resultsfile,$query_chain_opt,$gzipped);

#    print "ran alignments\n";
  }

	if($gzipped) {
	  foreach $target (@targets) {
	    $file = $alignment_dir."/".$query."_".$target.".out";
	    if(-e $file) {
	      system("$gzip -f $file");
	    }
	  }
	}

  #check to see if query vs. query alignment needs to be run
  my($query_vs_query) = $alignment_dir."/query_".$shortquery.".out";
  my($query_vs_query2) = $alignment_dir."/".$shortquery."_".$shortquery.".out";
  my($query_vs_query3) = $alignment_dir."/".$query_out_name."_".$query_out_name.".out";
  
  if($gzipped) {
    $query_vs_query .= ".gz";
    $query_vs_query2 .= ".gz";
    $query_vs_query3 .= ".gz";
  }
  if(! -e($query_vs_query) and ! -e($query_vs_query2) and ! -e($query_vs_query3)) {
    print "running $query vs $query alignment . . .\n";
    print "$lockprogram -q $query -t $query $query_chain_opt $target_chain_opt > /dev/null";
#    system("$lockprogram -q $query -t $query $query_chain_opt $target_chain_opt > /dev/null");
    system("$lockprogram -q $query -t $query -o alignments > /dev/null");

    if(! -e($alignment_dir)) {
      print STDERR "alignment dir ($alignment_dir) doesn't exist (running query alignment): creating\n";
      system("$mkdir $alignment_dir");
    }else {
      print STDERR "alignment dir does exist (running query alignment)\n";
    }
    if(-e "query_target.out") {
      system("$mv query_target.out $query_vs_query");
    }else {
#      my($file) = $query."_".$query.".out";

#      system("$mv $file $query_vs_query");
    }
    if($gzipped) {
      system("$gzip -f $query_vs_query");
    }
  }


  $num_alignments = scalar(@targets);


  #get appropriate background score distribution for the query structure
  $expectation_curve = getSCOPClassification($shortquery);
#  print "lambda=$lambda{$expectation_curve}, mu=$mu{$expectation_curve}\n";
  #determine which SSEs, if any, to exclude from motif detection
  %excluded=parseExcludeString($excludestring);

  #calculate statistical significance thresholds
  $pcutoff = $num_false_positives/$num_alignments;

  $stringent_pcutoff = $pcutoff/10.0;
  
  #read LOCK search.out file
  $num_query_SSEs=readSearchFile($shortquery,$resultsfile,$resultsdir,$alignment_dir);

  #motif discovery/results analysis
  @hits=runMotifDiscovery($shortquery,$num_query_SSEs,$xval,$expectation_curve,$pcutoff,$stringent_pcutoff,$resultsdir);

  #print out list of hits to text file
  printTextFileHits(\@hits,$resultsdir,"significant-results.txt");
  printHTMLTable(\@hits,$resultsdir,"significant-results.html");

  #print out sorted list of all alignments (including insignificant) to text file
  @allresults = sort(sort_by_pvalue keys(%pvalues)); #should this be sort_by_score, or should significant hits be sorted by pvalue?? (probably the latter -- need to change web page as well)
  printTextFileHits(\@allresults,$resultsdir,"all-results.txt");
  printHTMLTable(\@allresults,$resultsdir,"all-results.html");


  open(OUT, ">$resultsdir/$shortquery-SSE-conservations.txt");
  print OUT "Start\tEnd\tType\tConservation\n";
  for($i=0;$i<=$#query_SSE_starts;$i++) {
    printf(OUT "$query_SSE_starts[$i]\t$query_SSE_ends[$i]\t$query_SSE_types[$i]\t%.3f\n",$conservation{$query_SSE_starts[$i]});
  }
  close(OUT);

  #print rasmol script for motif colored by conservation
  my($high,$low,$colorstring,@color);
  $colorstring = "color gray\n";
  $high=0; $low=1;
  for($i=0;$i<=$#query_SSE_starts;$i++) {
    next if($excluded{$query_SSE_starts[$i]});
    $colorstring .= "select ($query_SSE_starts[$i]-$query_SSE_ends[$i] and not hetero)\n";
    @color = getColor($conservation{$query_SSE_starts[$i]});
    if($conservation{$query_SSE_starts[$i]} < $low) {
      $low = $conservation{$query_SSE_starts[$i]};
    }
    if($conservation{$query_SSE_starts[$i]} > $high) {
      $high = $conservation{$query_SSE_starts[$i]};
    }
    $colorstring .="color [$color[0],$color[1],$color[2]]\n";
  }
  $intervalsize = ($high-$low)/20.0;
  $ratio = (100/20)**(1/19);
  open(OUT,">$resultsdir/query_motif.spt");
  print OUT "background white\n";
  print OUT "cartoons on\nwireframe off\n";
  print OUT "$colorstring\n";
  close(OUT);

  if($make_pdb_files) {
    print "making pdb files\n";
    for($i=0;$i<=$#hits;$i++) {
      print "making pdb file for target $hits[$i]\n";
      $file = $alignment_dir."/$shortquery"."_".$hits[$i]."\.out";
      if($gzipped) {
	system("$gunzip -f $file.gz");
      }
      $error=system("./makePDBfile.pl $file > /dev/null");
      if($gzipped) {
	system("$gzip -f $file");
      }
    }
      
  }else {
#    print "not making pdb files:$make_pdb_files\n";
  }
  system("$rm -f .targetdb");
}


sub printUsage {

    print "Usage:\n";
    print "FoldMiner.pl -q <name of query structure> -t <target database file> [-a <alignment directory> -r <results directory> -e <expectation> -x <x value> -exclude <exclude string> -plg ]\n";
    print "\n";
    print <<END;
Explanations:
    -q: Name of query structure file (full path is not necessary); e.g. 3sdh.ent.Specify a single
        chain by appending a dash and the chain identifier (e.g. 3sdh-A)
    -t: Target database file. Each line must contain a SCOP identifier, a PDB identifier, or a 
        full path to a PDB file. To specify a single chain of a structure, append a dash
	and the chain identifier (e.g. chain A of 1dlw becomes 1dlw-A).
    -a: Path to directory where LOCK 2 alignments have been stored if
        alignments have already been run, or desired directory if 
        alignments have not been run. If ommitted, will be set to
        current directory
    -r: Path to search.out file if LOCK 2 alignments have already been run,
        or desired path to search.out file if alignments have not been run. May be
        excluded if alignment directory and results directory are the same.
    -e: Expectation. If not specified here or in foldminer.config, default value is 10
    -x: x value. If not specified here or in foldminer.config, default value is 0.75
    -exclude: Secondary structure elements to be excluded from analysis. A comma separated
              list of start residues, e.g. -exclude 8,25
    -p: Make PDB files for statistically significant alignments.
    -l: Optional argument. Specifies local (rather than global) alignments
    -g: Optional argument. Specifies that alignments are or should be gzipped
END
}

sub sort_by_score {
    my($domain1,$domain2) = ($a,$b);
    my($num1,$num2);
    $num1 = $dp_scores_significant{$domain1};
    $num2 = $dp_scores_significant{$domain2};

    $num2 <=> $num1;

}

sub sort_by_pvalue {
  my($domain1,$domain2) = ($a,$b);
  my($num1,$num2);
  $num1 = $pvalues{$domain1};
  $num2 = $pvalues{$domain2};
  $num1 <=> $num2;
}

sub inverse_expectation_survival_function {
  my($lambda,$mu,$expectation) = @_;
  my($temp);
  
  $temp = -log(log(1/(1-$expectation)));
  $temp/=$mu;
  return($temp+$lambda);
}

sub expectation_function {
  my($lambda,$mu,$x) = @_;
  return(1-exp(-exp(-$lambda*($x-$mu))));
}

sub inverse_exponential_function {
  my($x, $y) = @_;
  my($retvalue);

  return(-log($y)/$x);
}

#Given the conservation of an SSE, returns a color in rgb format. High conservations
#correspond to bright colors. The color scheme was developed in HSV colorspace with the
#brightest color corresponding to (353,95,80) and the darkest to (353,95,10). 
#Photoshop was used for RGB conversions. Colors are not necessarily "web safe," but
#should print on most printers.
sub getColor {
  my($c) = @_; #conservation

  if($c < 0.05) {
    return((26,1,4));
  }elsif($c < 0.1) {
    return((28,1,5));
  }elsif($c < 0.15) {
    return((31,2,5));
  }elsif($c < 0.2) {
    return((36,2,6));
  }elsif($c < 0.25) {
    return((38,2,6));
  }elsif($c < 0.3) {
    return((43,2,7));
  }elsif($c < 0.35) {
    return((48,2,8));
  }elsif($c < 0.4) {
    return((56,3,9));
  }elsif($c < 0.45) {
    return((61,3,10));
  }elsif($c < 0.5) {
    return((69,3,11));
  }elsif($c < 0.55) {
    return((77,4,12));
  }elsif($c < 0.6) {
    return((84,4,14));
  }elsif($c < 0.65) {
    return((94,5,15));
  }elsif($c < 0.7) {
    return((102,5,16));
  }elsif($c < 0.75) {
    return((117,6,19));
  }elsif($c < 0.8) {
    return((133,7,21));
  }elsif($c < 0.85) {
    return((148,7,24));
  }elsif($c < 0.9) {
    return((184,9,30));
  }else {
    return((204,10,33));
  }

}

sub printHTMLTable {
  my($resultsref,$resultsdir,$filename) = @_;
  my($i,$targetscoplink);
  my($fold,$class,$superfam,$fam);
  my($foldname);

  open(OUT,">$resultsdir/$filename") || die "can't open results.html for writing: $!\n";
  print OUT "<html>\n<body bgcolor=#FFFFFF>\n";
  print OUT "<h1>FoldMiner search results for query</h1>\n";
  

  print OUT "<table width=100% cellpadding=1 cellspacing=0 border=1>\n";
  print OUT "<tr>\n";
  print OUT "   <th align=center><b>Target</b></th>\n";
  print OUT "   <th align=center><b>Score</b></th>\n";
  print OUT "   <th align=center><b>P value</b></th>\n";
  print OUT "   <th align=center><b>Aligned SSEs</b></th>\n";
  print OUT "   <th align=center><b>Aligned Residues</b></th>\n";
  print OUT "   <th align=center><b>RMSD</b></th>\n";
  print OUT "   <th align=center><b>SCOP Fold</b></th>\n";
  print OUT "   <th align=center><b>PDB Header</b></th>\n";
  print OUT "</tr>\n";
  for($i=0;$i<=$#$resultsref;$i++) {
    $targetscoplink="http://scop.mrc-lmb.cam.ac.uk/scop/search.cgi?sid=$$resultsref[$i]";
    print OUT "<tr>\n<td align=left><a href=$targetscoplink target=$$resultsref[$i]>$$resultsref[$i]</a></td>\n";
    printf OUT ("<td align=center>%4.2f</td>\n",$finalscores{$$resultsref[$i]});
    printf OUT ("<td align=center>%1.2g</td>\n",$pvalues{$$resultsref[$i]});
    print OUT "<td align=center>$num_aligned_SSEs{$$resultsref[$i]}</td>\n";
    print OUT "<td align=center>$num_aligned_residues{$$resultsref[$i]}</td>\n";
    print OUT "<td align=center>$RMSD{$$resultsref[$i]}</td>\n";
    if(defined($classificaitons{$$resultsref[$i]})) {
	($class,$fold,$superfam,$fam) = split(/\./,$classifications{$$resultsref[$i]});
	$fold = $class."\.".$fold;
	$foldname = $foldnames{$fold};
    }else {
	$foldname = "N/A";
#	$fold = getSCOPClassification($$resultsref[$i]);
#	if($fold eq "all") {
#	    $foldname = "N/A";
#	}
    }
    print OUT "<td align=left><a href=\"http://scop.mrc-lmb.cam.ac.uk/scop/search.cgi?sccs=$fold\" target='$fold'>$foldname</a></td>\n";
    print OUT "<td align=left><font size=-1>$description{$$resultsref[$i]}</font></td>\n";
    print OUT "</tr>";
  }
  print OUT "</table>\n";
  print OUT "</body></html>";
  close(OUT);
}

sub printTextFileHits {
  my($resultsref,$resultsdir,$filename) = @_;
  my($i,$included);
  my($fold,$class,$superfam,$fam);
  my($foldname);

  open(OUT, ">$resultsdir/$filename") || die "can't open hits.txt for writing: $!\n";
  print OUT "Target\tScore\tp value\tAligned SSEs\tAligned residues\tSCOP Fold\tPDB Header\n";
  for($i=0;$i<=$#$resultsref;$i++) {
      if(defined($classifications{$$resultsref[$i]})) {
	  ($class,$fold,$superfam,$fam) = split(/\./,$classifications{$$resultsref[$i]});
	  $fold = $class."\.".$fold;
	  $foldname = $foldnames{$fold};
      }else {
	  $foldname = "N/A";
#	  $fold = getSCOPClassification($$resultsref[$i]);
#	  if($fold eq "all") {
#	      $foldname = "N/A";
#	  }
      }
    printf OUT ("$$resultsref[$i]\t%.2f\t%.2g\t$num_aligned_SSEs{$$resultsref[$i]}\t$num_aligned_residues{$$resultsref[$i]}\t$RMSD{$$resultsref[$i]}\t$foldname\t$description{$$resultsref[$i]}\n",$finalscores{$$resultsref[$i]},$pvalues{$$resultsref[$i]});
  }
  close(OUT);
}
