#!/usr/bin/perl
# Copyright 2003 The Board of Trustees of
# The Leland Stanford Junior University. All Rights Reserved.

#Please send comments, questions, and bug reports to Jessica Shapiro (jessicas@stanford.edu)


# This script opens search.out and calculates the statistical significance of the alignment
# using background score distributions specific to the query's SCOP fold, if known. Note
# that LOCK 2 prints out statistical significance values using a background score distribution
# covering all SCOP folds. Your search.out file will be replaced with one that contains
# the new, more accurate p values.

# Usage: ./calculate_pvalues.pl <name of query structure OR SCOP fold identifier> <path to search.out>

# Usage Notes: 
# 1) The name of the query structure is its seven character SCOP id. If your query
#    structure is not a SCOP domain, it is not useful to run this script with the name of
#    the query structure specified. Alternatively, you may enter an actual SCOP fold 
#    (e.g. "a.1"). If your query is not a SCOP domain but you wish to calculate p values
#    using a background score distribution for a specific fold, this is an useful option.
#
# 2) LOCK 2 prints out a search.out file each time it is run. You may rename this file. This
#    script expects the full path (e.g. mylockruns/experiment1.out or mylockruns/search.out).
#    The file will be replaced with a file of the same name containing the newly calculated 
#    p values.



use lib "/var/www/cgi-bin/FoldMiner2";

use Expectation qw(%lambda %mu); #mu and lambda are hashes containing expectation curve 
                                  #parameters for SCOP folds. Keys are SCOP 
                                  #folds (e.g. "a.1")
use Scop qw(%classifications); #%classifications gives a SCOP domain's full classification (e.g. a.1.1.1)


if($#ARGV != 1) {
  print STDERR "Usage: ./calculate_pvalues.pl <query SCOP id OR SCOP fold identifier> <path to search.out>\n";
  print STDERR" (See comments at top of calculate_pvalues.pl for detailed instructions.)\n";
  exit(0);
}

$query = shift(@ARGV); #the query structure's name (important only if SCOP id)
$resultsfile = shift(@ARGV); #the search.out file produced by LOCK 2

if($query =~ /^[A-Za-z]\.[0-9]+$/) { #scop fold was specified
  $background_dist = $query;
  $background_dist =~ tr/A-Z/a-z/;
  if(! defined($lambda{$background_dist})) {
    $queryclass = substr($query,0,1);
    if(defined($labmda{$queryclass})) {
      $background_dist=$queryclass;
    }else {
      $background_dist = "all";
    }
  }
}else {
  #Get the SCOP fold of the query, if available, and set the background distribution.
  
  if(!defined($classifications{$query})) {
    $background_dist = "all"; #could use bg distribution for the structure's class
  }else {
    ($class,$fold,$superfam,$fam) = split(/\./,$classifications{$query});
    $background_dist = $class.".".$fold;
    if(!defined($lambda{$background_dist})) {
      if(defined($lambda{$class})) {
	$background_dist =  $class;
      }else {
	$background_dist = "all";
      }
    }

  }
}


open(IN, $resultsfile) || die "can't open $resultsfile for reading: $!\n";
@lines = <IN>;
close(IN);
open(OUT, ">$resultsfile") || die "can't open $resultsfile for writing: $!\n";

for $line (@lines) {
  @fields = split(/\|/,$line);
  $score = $fields[2]; #this is the normalized alignment score

  $pval = expectation_function($lambda{$background_dist},$mu{$background_dist},$score);
  $fields[1] = $pval;
  printf(OUT "$fields[0]|%.4g|$fields[2]|$fields[3]|$fields[4]|$fields[5]|$fields[6]",$pval);
  for($i=7;$i<=$#fields;$i++) {
    print OUT "|$fields[$i]";
  }
  print OUT "\n";
}
close(OUT);

#Computes the value of the extreme value distribution with parameters $lambda and $mu at the value $x
sub expectation_function {
  my($lambda,$mu,$x) = @_;

  return(1-exp(-exp(-$lambda*($x-$mu))));
}

